/******************************************************************************/
/* $Header: /cvsroot/umbrella/umbrella-devel/kernel/security/umbrella/umb_types.c,v 1.22 2005/03/16 10:23:24 mthrysoe Exp $ */
/******************************************************************************/
/*  AUTHOR:                                                                   */
/******************************************************************************/
/*                                                                            */
/*    Sren Nhr Christensen                                                  */
/*    Kristian Srensen                                                       */
/*    Michel Thryse                                                          */
/*    mailto:umbrella@cs.auc.dk                                               */
/*    http://umbrella.sourceforge.net                                         */
/*                                                                            */
/******************************************************************************/
/*  COPYRIGHT:                                                                */
/******************************************************************************/
/*                                                                            */
/*    Copyright (c) 2004 The Umbrella Security Team.                          */
/*    All rights reserved.                                                    */
/*                                                                            */
/******************************************************************************/
/*  LICENSE:                                                                  */
/******************************************************************************/
/*                                                                            */
/*    This file is part of the Umbrella Security Project and is free          */
/*    software; you can redistribute it and/or modify it under the terms      */
/*    of the GNU Library General Public License as published by the Free      */
/*    Software Foundation;                                                    */
/*    The code is distributed in the hope that it will be useful,             */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        */
/*    Library General Public License for more details.                        */
/*                                                                            */
/*    You should have received a copy of the GNU Library General Public       */
/*    License along with this library; if not, write to the                   */
/*    Free Software Foundation, Inc.,                                         */
/*    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                            */
/*    or download a copy from ftp://ftp.gnu.org/pub/gnu/COPYING.LIB-2.0       */
/*                                                                            */
/******************************************************************************/

#include <linux/security.h>
#include "include/umb_types.h"
#include "include/umb_bitvector.h"
#include "include/umb_signature.h"

/******************************************************************************/
/* UMBRELLA TYPES: security struct memory management functions                */
/******************************************************************************/

struct security_struct *security_struct_create() {
	struct security_struct* security;
	int i = 0;

	security = (struct security_struct *) kmalloc(sizeof(struct security_struct), GFP_ATOMIC);

	if (!security)
		return NULL;

	security->level1 = bv_create();
	security->child_level1 = bv_create();
	security->level2 = init_fsr();

	for (i = 0; i<MAX_L2; i++)
		security->child_level2[i] = NULL;


	/* extra fields in security struct for holding binary restrictions */
	security->bin_level1 = bv_create();

	for (i = 0; i<MAX_L2; i++)
		security->bin_level2[i] = NULL;

	if (!security->level1 || !security->child_level1 || !security->level2 || 
		!security->bin_level1 || !security->bin_level2 ) {
		printk("Umbrella: security_struct_create: Not enough memory!\n");
		return NULL;	
	}	
	return security;
}

void security_struct_destroy(struct security_struct *target){
	int i = 0;

	while (target->child_level2[i] != NULL)
		kfree(target->child_level2[i++]);

	while (target->bin_level2[i] != NULL)
		kfree(target->bin_level2[i++]);
	
	bv_destroy(target->bin_level1);
	bv_destroy(target->level1);
	bv_destroy(target->child_level1);
	destroy_fsr_tree(target->level2);
}

struct fsr *init_fsr() {
	struct fsr *fsr = (struct fsr *) kmalloc(sizeof(struct fsr), GFP_ATOMIC);
	fsr->succ = NULL;
	fsr->next = NULL;
	fsr->name = NULL;

	return fsr;
}

int insert_into_fsr(struct fsr *root, char **path) {	
	struct fsr *current_fsr = root;
	
	while (1) {
		if (*path == NULL) {
			if (current_fsr->succ == NULL) {
				//printk("Restrction exists!\n");
				return SUCCES;
			}
			
 			if (current_fsr->succ != NULL) {
				destroy_fsr_tree(current_fsr); 
				//printk("Restrction exists, freed subtree!\n");
				return SUCCES;
			}
		}
		else if (current_fsr->name != NULL && !strcmp(current_fsr->name, *path)) {
			if (current_fsr->succ == NULL) {
				//printk("fsr->name: %s, inserting: %s in fsr->succ\n", current_fsr->name, *(++path));
				current_fsr->succ = create_fsr_subtree(path);
				return SUCCES;
			}
			if (current_fsr->succ != NULL) {
				current_fsr = current_fsr->succ;
				*path++;
				//printk("Traversing downwards.!\n");
				
			}	
		}
		else if (current_fsr->next == NULL) {
			//printk("fsr->name: %s, inserting: %s in fsr->next\n", current_fsr->name, *path);
			current_fsr->next = create_fsr_subtree(path);
			return SUCCES;
		}
		else
			current_fsr = current_fsr->next;
	} 
}

struct fsr *create_fsr_subtree(char **path) {
	struct fsr *new = NULL;
	if (*path != NULL) {
		new = (struct fsr *) kmalloc(sizeof(struct fsr), GFP_ATOMIC);
		new->name = *path;
		new->next = NULL;
		new->succ = create_fsr_subtree(++path);
	}
	return new;
}

struct fsr *copy_fsr_tree(struct fsr *source) {
	struct fsr *new = (struct fsr *) kmalloc(sizeof(struct fsr), GFP_ATOMIC);

	if (source->name == NULL)
		new->name = NULL;
	else {
		new->name = (char *) kmalloc(sizeof(char)*(strlen(source->name)+1), GFP_ATOMIC);
	strcpy(new->name, source->name);
	}
		
	if(source->succ != NULL) {
		new->succ = copy_fsr_tree(source->succ);
	} else
		new->succ = NULL;
	
	if(source->next != NULL) {
		new->next = copy_fsr_tree(source->next);
	}
	else
		new->next = NULL;
	return new;
}

void destroy_fsr_tree(struct fsr *input) {
	struct fsr *tmp_next, *tmp_succ;

	tmp_succ = input->succ;
	tmp_next = input->next;
	kfree(input->name);
	kfree(input);
	if(tmp_succ != NULL)
		destroy_fsr_tree(tmp_succ);
	if(tmp_next != NULL)
		destroy_fsr_tree(tmp_next);
	return;
}

int check_restriction_in_fsr(struct fsr *root, char **path) {
	struct fsr *fsr = root;
	int retval = -1;
		
	if (*path == NULL)
		return ALLOWED;

	if (fsr->name == NULL && fsr->succ == NULL) {
		fsr = fsr->next;
	}

	if (!fsr) {
		return ALLOWED;
	}
	
	if (retval == -1 && !strcmp(*path, fsr->name)) {
		
		if (fsr->succ == NULL) {
			//printk("Denying due to no succ and mathcing path\n");
			retval = DENIED;
		}
		else {
			if (++path != NULL) {
				path--;
				//printk("Recursive call on %s ->succ with %s\n", fsr->name, *path);
				retval = check_restriction_in_fsr(fsr->succ, ++path);
			} else
				retval = ALLOWED;
		}

	} else if (retval == -1 && strcmp(*path, fsr->name)) {
		
		if (fsr->next != NULL) {
			//printk("Recursive call on %s ->next with %s\n", fsr->name, *path);
			retval = check_restriction_in_fsr(fsr->next, path);
		} else
			retval = ALLOWED;
	}
	return retval;
		
}
