/***************************************************************************/
/* $Header: /cvsroot/umbrella/umbrella-devel/kernel/security/umbrella/umbrella.c,v 1.87 2005/03/16 10:23:24 mthrysoe Exp $ */
/***************************************************************************/
/*  AUTHOR:                                                                */
/***************************************************************************/
/*                                                                         */
/*    Sren Nhr Christensen                                               */
/*    Kristian Srensen                                                    */
/*    Michel Thryse                                                       */
/*    mailto:umbrella@cs.aau.dk                                            */
/*    http://umbrella.sourceforge.net                                      */
/*                                                                         */
/***************************************************************************/
/*  COPYRIGHT:                                                             */
/***************************************************************************/
/*                                                                         */
/*    Copyright (c) 2004 The Umbrella Team.                                */
/*    All rights reserved.                                                 */
/*                                                                         */
/***************************************************************************/
/*  LICENSE:                                                               */
/***************************************************************************/
/*                                                                         */
/*    This file is part of the Umbrella Security Project and is free       */
/*    software; you can redistribute it and/or modify it under the terms   */
/*    of the GNU Library General Public License as published by the Free   */
/*    Software Foundation;                                                 */
/*    The code is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU     */
/*    Library General Public License for more details.                     */
/*                                                                         */
/*    You should have received a copy of the GNU Library General Public    */
/*    License along with this library; if not, write to the                */
/*    Free Software Foundation, Inc.,                                      */
/*    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                */
/*                                                                         */
/*    or download a copy from ftp://ftp.gnu.org/pub/gnu/COPYING.LIB-2.0    */
/*                                                                         */
/***************************************************************************/

#include <linux/types.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/security.h>
#include <linux/gfp.h>
#include <linux/linkage.h>
#include <linux/errno.h>
#include <linux/namei.h>
#include <linux/socket.h>

#include <linux/proc_fs.h>
#include <linux/file.h>
#include <asm/uaccess.h>

#include "include/umb_bitvector.h"
#include "include/umb_types.h"
#include "include/umb_signature.h"
#include "include/umb_keyring.h"


/****************** Umbrella constants ***************************/

#if defined(CONFIG_SECURITY_UMBRELLA_MODULE)
  #define MY_NAME THIS_MODULE->name
#else
  #define MY_NAME "Umbrella"
#endif

static struct proc_dir_entry *umb_proc_entry;


MODULE_DESCRIPTION("Umbrella - Process based access control "
                   "and authentication of binaries");
MODULE_LICENSE("GPL");
  
int secondary;

int pid_to_apply;

/**************** Umbrella system calls **************************/

static void chop_string(char **output, char *input) {
    
    int count = 0;
    int i;
    int j = 1;
    int length = strlen(input);
    char *buffer;
	char *slash;
	
	slash = (char *)kmalloc(sizeof(char)*2, GFP_ATOMIC);
	if (!slash) {
		printk(KERN_ERR
			   "Umbrella ERROR: Not enough memory to allocate slash buffer\n");
		return;
	}

	slash[0] = '/';
	slash[1] = '\0';	
	
	output[0] = slash;
	
    while(input[count] != '\0') {
		buffer = (char *)kmalloc(sizeof(char)*length, GFP_ATOMIC);
		if (!buffer) {
			printk(KERN_ERR 
				   "Umbrella ERROR: Not enough memory to allocate buffer\n");
			return;
		}

		for(i=0;i<length;i++)
			buffer[i] = '\0';
		
		count++;
		i = 0;
		
		while(input[count] != '/' && input[count] != '\0') {
			buffer[i++] = input[count++];
		}
		output[j++] = buffer; 
    }
}

void umb_set_child_restrictions(int level1[], char **level2) {

	int i = 0;
	struct security_struct *security = current->security;
	
	/* -10 is termination symbol of level 1 restrictions */
	for (i=0; i<NUMBER_OF_LEVEL1; i++) {
		if (level1[i] == -10)
			break;
		bv_bit_on(security->child_level1, level1[i]);
	}
	
	if((level2 != NULL) && (*level2 != NULL)) {
		bv_bit_on(security->child_level1, 0);
		for (i=0; i<MAX_L2; i++) {
			if (level2[i] == NULL) break;
			security->child_level2[i] =
				(char *)kmalloc(sizeof(char) * (strlen(level2[i]) + 1), GFP_ATOMIC);

			/* handle not enough memory situation */
			if (!security->child_level2[i]) {
				while (i>0) {
					kfree(security->child_level2[i]);
					i--;
				}
				printk(KERN_ERR "UMBRELLA: Not enough memory for setting "
					            "child's restrictions\n");
				return;
			}
			
			strcpy(security->child_level2[i], level2[i]);
		}
	}
	return;
}

/******************** Umbrella helper functions *************************/


static int length_is(char **array) {
	int i;

	for(i=0; i<MAX_FILE_LENGTH; i++) {
		if (array[i] != NULL)
			continue;
		return i;
	}
 
	return -EOVERFLOW;
}

static void reverse(char **array, int length) {
	int i, j;
	char *tmp;
	
	for (i=0, j=length-1; i<(length>>1); i++, j--)
		{
			tmp = array[i];
			array[i] = array[j];
			array[j] = tmp;
		}
	
	return;
}

/* parses the path of an accessed file dentry */
int parse_path(struct dentry *d, char **path_array) {
	int done = 0;
	int count = 0;
	int length;

	/* fill path in path_array */
	while(!done) {
		if(count == (MAX_PATH_DEPTH-1)) {
			printk(KERN_ERR "Umbrella umb_parse_path: Overflow in path_depth\n");
			return -EOVERFLOW;
		}

		path_array[count] = (char *)d->d_name.name;

		if(!(strcmp(path_array[count], "/")))
			done = 1;
		else if(!(strcmp(path_array[count], "pipe:")) ||
			   (!(strcmp(path_array[count], "socket:"))))
				   done = 1;
		else {			
			d = d->d_parent;
			count++;
		}
	}	
	length = length_is(path_array);

	if (length == -EOVERFLOW) {
		printk(KERN_ERR "Umbrella umb_parse_path: Overflow in length of path!\n");
		return -EOVERFLOW;
	}

	reverse(path_array, length);
	return 0;
}

static inline int file_hook_wrapper(struct dentry *dentry) {
	int ss_decision, i = 0;
	char *path[MAX_PATH_DEPTH];
	struct security_struct *current_security = current->security;

	/* initialize path array, not sure if this is needed, snc */
	for (i = 0; i < MAX_PATH_DEPTH; i++)
		path[i] = NULL;
	/* TODO OPTIMIZE ME: this can be done in parse_path - then only ONE null */
	/*              is needed to be assigned - i.e. this loop can be avoided */

	/* Parsing the path */
	if (parse_path(dentry, path) == -EOVERFLOW)
		ss_decision = -EOVERFLOW;
	else {
		/* Getting the decision */
		ss_decision = check_restriction_in_fsr(current_security->level2, path);
		if (ss_decision)
			ss_decision = -EPERM;
	}
	return ss_decision;
}

/* OLD */
/* void umb_parse_proc(char *string) { */

/* 	char buffer[UMB_POLICY_SIZE]; */
/* 	char *tmpfsr[MAX_L2]; */
/* 	char *tmpstr; */
/* /\* 	char *id; *\/ */
/* 	bitvector tmpnfsr; */
/* 	int sigpos, bufpos; */
/* 	int i, end, counter, loop_counter; */

/* 	if (strlen(string) != UMB_POLICY_SIZE) { */
/* 		printk("Umbrella: Error parsing /proc/umbrella - wrong policy size!\n"); */
/* 		return; */
/* 	} */


/* 	/\* VERIFY BEGIN TAG *\/ */
/* 	for (i = 0; i < UMB_POLICY_SIZE; i++) buffer[i] = '\0'; */

/* 	sigpos = 0; */
/* 	bufpos = 0; */
/* 	end = UMB_POLICY_BEGIN_TAG; */
/* 	while (sigpos < end) */
/* 		buffer[bufpos++] = string[sigpos++]; */

/* 	if (!strcmp(buffer, "#umbrella_start#")) { */
/* 		printk("Umbrella: Warning: Policy in /proc/umbrella not valid!\n"); */
/* 		return; */
/* 	} */
/* 	printk("Buffer 001 is: %s\n", buffer); */

/* 	return; */
		
/* 	/\* VERIFY ID TAG *\/ */
/* 	for (i = 0; i < UMB_POLICY_BEGIN_TAG; i++) buffer[i] = '\0'; */

/* 	bufpos = 0; */
/* 	end = end + UMB_POLICY_ID_TAG; */
/* 	while (sigpos < end) */
/* 		buffer[bufpos++] = string[sigpos++]; */

/* 	if (!strcmp(buffer, "#id:")) { */
/* 		printk("Umbrella: Warning: Policy in /proc/umbrella not valid!\n"); */
/* 		return; */
/* 	} */

/* 	/\* FETCH ID *\/ */
/* 	for (i = 0; i < UMB_POLICY_ID_TAG; i++) buffer[i] = '\0'; */

/* /\* **** The commented region below can be used in umb_parse_sig *\/ */
/* /\* 	bufpos = 0; *\/ */
/* 	end = end + UMB_POLICY_ID_SIZE; */
/* 	sigpos = sigpos + UMB_POLICY_ID_SIZE; /\* Only enable when other code is commented out!! *\/ */
/* /\* 	while (sigpos < end && strcmp(string[sigpos], ':')) *\/ */
/* /\* 		buffer[bufpos++] = string[sigpos++]; *\/ */

/* /\* 	id = (char *) kmalloc(sizeof(char)*strlen(buffer), GFP_ATOMIC); *\/ */
/* /\* 	if (!fsr) { *\/ */
/* /\* 		printk("Umbrella: Critical: Memory allocation failed parsing /proc/umbrella\n"); *\/ */
/* /\* 		return; *\/ */
/* /\* 	} *\/ */
/* /\* 	strcpy(id, buffer); *\/ */

/* 	/\* VERIFY NFSR TAG *\/ */
/* 	for (i = 0; i < UMB_POLICY_ID_SIZE; i++) buffer[i] = '\0'; */

/*     /\* sigpos hack as the id may be shorter than id size *\/ */
/* /\* 	sigpos = sigpos - bufpos + UMB_POLICY_ID_SIZE; *\/ /\* Only enable if above block is active! *\/ */
/* 	bufpos = 0; */
/* 	end = end + UMB_POLICY_NFSR_TAG; */
/* 	while (sigpos < end) */
/* 		buffer[bufpos++] = string[sigpos++]; */

/* 	if (!strcmp(buffer, "#l1:")) { */
/* 		printk("Umbrella: Warning: Policy in /proc/umbrella not valid!\n"); */
/* 		return; */
/* 	} */

/* 	/\* FETCH NFSR *\/ */
/* 	for (i = 0; i < UMB_POLICY_ID_SIZE; i++) buffer[i] = '\0'; */

/* 	tmpnfsr = bv_create(); */
/* 	if (!tmpnfsr) { */
/* 		printk("Umbrella: Critical: Memory allocation failed parsing /proc/umbrella\n"); */
/* 		return; */
/* 	} */

/* 	bufpos = 0; */
/* 	loop_counter = 0; */
/* 	end = end + UMB_POLICY_NFSR_SIZE; */
/* 	while (sigpos < end) { */
/* 		if (string[sigpos] == (char) ':' && strlen(buffer) > 0) { */
/* 			/\* see "man ctoi" for strtol explanation *\/ */
/* 			bv_bit_on(tmpnfsr, simple_strtol(buffer, (char **)NULL, 10)); */
/* 			buffer[0] = '\0'; /\* we only use the first two as the vector is 32 bit *\/ */
/* 			buffer[1] = '\0'; */
/* 			bufpos = 0; */
/* 		} */
/* 		else { */
/* 			if (string[sigpos] == (char) ':' && string[sigpos+1]  == (char) ':') { */
/* 				printk("Umbrell Debug: No more nfsr - two succetive :'s occured\n"); */
/* 				sigpos = sigpos + (UMB_POLICY_NFSR_SIZE - loop_counter); */
/* 				break; */
/* 			} */
/* 			buffer[bufpos++] = string[sigpos]; */
/* 		} */
/* 		sigpos++; */
/* 		loop_counter++; */
/* 	} */
	
/* 	/\* VERIFY FSR TAG *\/ */
/* 	for (i = 0; i < UMB_POLICY_NFSR_SIZE; i++) buffer[i] = '\0'; */

/* 	bufpos = 0; */
/* 	end = end + UMB_POLICY_FSR_TAG; */
/* 	while (sigpos < end) */
/* 		buffer[bufpos++] = string[sigpos++]; */

/* 	if (!strcmp(buffer, "#l2:")) { */
/* 		printk("Umbrella: Warning: Policy in /proc/umbrella not valid!\n"); */
/* 		bv_destroy(tmpnfsr); */
/* 		return; */
/* 	} */

/* 	/\* FETCH FSR *\/ */
/* 	for (i = 0; i < UMB_POLICY_FSR_TAG; i++) buffer[i] = '\0'; */

/* 	bufpos = 0; */
/* 	counter = 0; */
/* 	loop_counter = 0; */
/* 	end = end + UMB_POLICY_FSR_SIZE; */
/* 	while (sigpos < end) { */
/* 		if (string[sigpos] == (char) ':' && strlen(buffer) > 0) { */
/* 			tmpstr = (char *) kmalloc(sizeof(char)*strlen(buffer), GFP_ATOMIC); */
/* 			if (!tmpstr) { */
/* 				printk("Umbrella: Critical: Memory allocation failed parsing /proc/umbrella\n"); */
/* 				for (i = 0; i < MAX_L2; i++) kfree(tmpfsr[i]); */
/* 				bv_destroy(tmpnfsr); */
/* 				return; */
/* 			} */
/* 			strcpy(tmpstr, buffer); */
/* 			tmpfsr[counter] = tmpstr; */
/* 			for (i = 0; i < UMB_POLICY_FSR_SIZE; i++) buffer[i] = '\0'; */
/* 			bufpos = 0; */
/* 			counter++; */
/* 			sigpos++; */
/* 		} */
/* 		else { */
/* 			if (string[sigpos] == (char) ':' && string[sigpos+1]  == (char) ':') { */
/* 				printk("Umbrell Debug: No more fsr - two succetive :'s occured\n"); */
/* 				sigpos = sigpos + (UMB_POLICY_FSR_SIZE - loop_counter); */
/* 				break; */
/* 			} */
/* 			buffer[bufpos++] = string[sigpos]; */
/* 		} */
/* 		loop_counter++; */
/* 	} */

/* 	/\* VERIFY END TAG *\/ */
/* 	for (i = 0; i < UMB_POLICY_FSR_SIZE; i++) buffer[i] = '\0'; */
	
/* 	bufpos = 0; */
/* 	end = end + UMB_POLICY_END_TAG; */

/* 	while (sigpos < end) */
/* 		buffer[bufpos++] = string[sigpos++]; */
	
/* 	if (!strcmp(buffer, "#umbrella_end#")) { */
/* 		printk("Umbrella: Warning: Policy in /proc/umbrella not valid!\n"); */
/* 		bv_destroy(tmpnfsr); */
/* /\* 		kfree(fsr); *\/ */
/* 		return; */
/* 	} */


/* 	/\* EVERYTHING IS OKAY - COPY POLICY TO CURRENT->SECURITY->CHILD-* *\/ */

/* } */


int get_chartype(char c) {

    int retval = -1;

    switch(c) {
		case '0' : retval = IS_INT;   break;
		case '1' : retval = IS_INT;   break;
		case '2' : retval = IS_INT;   break;
		case '3' : retval = IS_INT;   break;
		case '4' : retval = IS_INT;   break;
		case '5' : retval = IS_INT;   break;
		case '6' : retval = IS_INT;   break;
		case '7' : retval = IS_INT;   break;
		case '8' : retval = IS_INT;   break;
		case '9' : retval = IS_INT;   break;
		case '/' : retval = IS_SLASH; break;
		case ':' : retval = IS_COLON; break;
    }
    
    return retval;

}

void umb_parse_and_apply_proc(char *policy) {

	struct security_struct *security = current->security;
    int i, j, len, tmpindex, no_fsr;
    char buffer[1024];
    char *tmpfsr[MAX_L2];
	char *path;
    bitvector tmpnfsr = bv_create();

#ifdef CONFIG_SECURITY_UMBRELLA_DEBUG
	printk(KERN_DEBUG "Umbrella policy loaded from /proc/umbrella is %s\n", policy);
#endif

	if (!tmpnfsr) {
		printk(KERN_ERR "Umbrella ERROR: Could not allocate memory for bitvector\n");
		return;
	}

	for (i = 0; i < MAX_L2; i++)
		tmpfsr[i] = NULL;

    len = strlen(policy);
    i = 0;
	no_fsr = 0;

    while (i < len) {
		switch (get_chartype(policy[i])) {

			case IS_INT :
				for (j = 0; j < 1024; j++)
					buffer[j] = '\0';

				j = 0;
				while (get_chartype(policy[i]) == IS_INT && i < len)
					buffer[j++] = policy[i++];

				tmpindex = simple_strtol(buffer, (char **)NULL, 10);
				if (tmpindex > 31) /* 32 bit bitvector only! */
					printk(KERN_INFO "Umbrella warning: Index out of bounce "
                           "in /proc/umbrella - ignoring!\n");
				bv_bit_on(tmpnfsr, tmpindex);

				break;


			case IS_SLASH :
				for (j = 0; j < 1024; j++)
					buffer[j] = '\0';

				j = 0;
				while (get_chartype(policy[i]) != IS_COLON && i < len)
					buffer[j++] = policy[i++];

				tmpindex = strlen(buffer);
				path = (char *) kmalloc(sizeof(char) * tmpindex, GFP_ATOMIC);

				if (!path) {
					printk(KERN_ERR "Umbrella ERROR: Could not allocate "
						   "memory for path in policy\n");
					bv_destroy(tmpnfsr);
					i = 0; while (tmpfsr[i] != NULL) kfree(tmpfsr[i++]);
					return;
				}

				for (j = 0; j < tmpindex; j++)
					path[j] = '\0';

				strcpy(path, buffer);
				tmpfsr[no_fsr++] = path;

				break;
				

			case IS_COLON :
				i++;
				break;


			default :
				printk(KERN_INFO "Umbrella ERROR: Invalid policy in "
                                 "/proc/umbrella\n");
				return;
		}
	}

	/* everything went okay - we not apply the restrictions */

#ifdef CONFIG_SECURITY_UMBRELLA_DEBUG
	printk(KERN_DEBUG "Umbrella DEBUG: proc applying to pid %i\n",
		   current->pid);
#endif

	bv_copy(security->child_level1, tmpnfsr);

	for (i = 0; i < MAX_L2; i++)
		security->child_level2[i] = tmpfsr[i];

	return;
}


static int umb_proc_write(struct file *file, const char *buffer,
                          unsigned long count, void *data) {

	char *policy;
	int i;

	policy = (char *)kmalloc(sizeof(char)*count, GFP_ATOMIC);
	if (!policy) {
		kfree(policy);
		return -EFAULT;
	}
	for (i = 0; i < count; i++) policy[i] = '\0';
	if (copy_from_user(policy, buffer, count)) {
		kfree(policy);
		return -EFAULT;
	}

#ifdef CONFIG_SECURITY_UMBRELLA_DEBUG
	printk(KERN_DEBUG "Umbrella DEBUG: policy is: %s\n", policy);
#endif

	umb_parse_and_apply_proc(policy);
	kfree(policy);

	return count;
}


/******************* Umbrella hookfunctions ***************************/

/* Hooks for handeling binary files - Start */

int umb_bprm_check_security (struct linux_binprm * bprm) {
	
	return  umb_handle_signature(bprm);
}

/* Hooks for handeling binary files - End */

static int umb_inode_create (struct inode *dir, struct dentry *dentry, int mode) {
	return file_hook_wrapper(dentry);
}

static int umb_inode_permission (struct inode *inode, int mask,
                                 struct nameidata *nd) {
	/* nd is NULL very often - if not checked, kernel will crash */
	if (nd == NULL)	return 0;

#ifdef CONFIG_SECURITY_UMBRELLA_DEBUG
	/* register PID that writes to /proc/umbrella */
	if (!strcmp((char *)nd->dentry->d_name.name, "umbrella") &&
        !strcmp((char *)nd->dentry->d_parent->d_name.name, "/")) {
		
		pid_to_apply = current->pid;
		printk(KERN_DEBUG "Umbrella DEBUG: inode_perm pid to apply is %i\n",
			   pid_to_apply);
	}
#endif

	return file_hook_wrapper(nd->dentry);
}

static int umb_inode_unlink(struct inode *dir, struct dentry *dentry) {
	return file_hook_wrapper(dentry);
}

static int umb_inode_mkdir (struct inode *dir, struct dentry *dentry, int mode) {
	return file_hook_wrapper(dentry);
}

static int umb_inode_link (struct dentry *old_dentry,
				       struct inode *dir,
				       struct dentry *new_dentry) {
	return file_hook_wrapper(old_dentry);
}

static int umb_inode_rename(struct inode *old_dir, struct dentry *old_dentry,
                            struct inode *new_dir, struct dentry *new_dentry) {
  return ((file_hook_wrapper(old_dentry) || file_hook_wrapper(new_dentry)) ?
          -EPERM : 0);
}

static int umb_inode_setattr (struct dentry *dentry,
										  struct iattr *attr) {
	return file_hook_wrapper(dentry);
}

static int umb_task_create(unsigned long clone_flags) {
	int ss_decision = 0;
	struct security_struct *security = current->security;

	/* current security is null when the system is booting */
	/* this line is inserted to prevent a kernel crash */
	if (current->security == NULL)
		return 0;
	
	ss_decision = bv_testbit(security->level1, NOFORK);

	if (ss_decision == 1 || ss_decision == -1)
		ss_decision = -EPERM;

	return ss_decision;
}

static int umb_socket_create(int family, int type, int protocol, int
kern) {
        int ss_decision = 0;
        struct security_struct *security = current->security;
		
        switch (family) {
        case 2:
			ss_decision = bv_testbit(security->level1, NOIP);
          break;

        case 23:
			ss_decision = bv_testbit(security->level1, NOIRDA);
          break;

        case 31:
			ss_decision = bv_testbit(security->level1, NOBLUETOOTH);
		  break;

	 	case PF_AX25:
          printk(KERN_DEBUG "This is the case for testing ;-)\n");
 		  break;  
		}
        
        if (ss_decision == 1 || ss_decision == -1)
          ss_decision = -EPERM;

        return ss_decision;
}


static int umb_task_alloc_security(struct task_struct *p) {
  struct security_struct *security = security_struct_create();
  struct security_struct *current_security = current->security;
  struct fsr *level2;       
  char **tmpl2;
  char **tmpl3;

  char *tmp[256];
  int i, ret = 0;

  /* return error if no security struct was created */
  if (!security)
	  return 1;

  for(i=0;i<256;i++)
	  tmp[i] = NULL;

  /* check if current is init */
  if (current_security != NULL) {
	  /* level 1 */
	  bv_or((bitvector)security->level1,
			(bitvector)current_security->child_level1, /* additional restr */
			(bitvector)current_security->level1);      /* from parent      */

	  bv_reset(current_security->child_level1); /* reset child restrictions */

	  /* level 2 inheritance */
	  level2 = copy_fsr_tree(current_security->level2);

	  /* insert additional restrictions into fsr tree */
      tmpl2 = current_security->child_level2;
	  i = 0;

	  while (tmpl2[i] != NULL && i < MAX_L2) {
		  chop_string(tmp, tmpl2[i]);
		  if (insert_into_fsr(level2, tmp)) {
			  printk(KERN_INFO "Umbrella: task_alloc_security: ");
			  printk(KERN_INFO "Failed to insert into hash_table!\n");
			  ret = 1; /* Failed */
			  /* TODO FIX ME: handle mem failure - break from while etc... */
		  }

		  kfree(tmpl2[i]);
		  tmpl2[i] = NULL;
		  i++; 
      }

	  /* if any restricitons was found on the binary in bprm_check_security
	     the are transfered to the new resulting process below. The values found
		 in the bin_levelX fields originate from the hook bprm_check_security, 
		 since it is called just before task_alloc_security and no easy link 
		 between the two exist elsewhere.
	  */
	  
	  /* first the level 1 restrictions */

	  if ( current_security->bin_level1 != NULL) {
		  bv_or((bitvector)security->level1,
				(bitvector)security->level1,  /* new p's restr */
				(bitvector)current_security->bin_level1); /* from binary */  
	  }   
 	    
	  bv_reset(current_security->bin_level1); /* binary restrictions already set */
  
	  /* then the fsr restrictions */

 	  if (current_security->bin_level2 != NULL) {
		  tmpl3 = current_security->bin_level2;
		  
	  	  /* insert additional restrictions into fsr tree */
		  i = 0;
		  while (tmpl3[i] != NULL) {
			  chop_string(tmp, tmpl3[i]);
			  if (insert_into_fsr(level2, tmp)) {
				  printk(KERN_INFO "Umbrella: task_alloc_security: ");
				  printk(KERN_INFO "Failed to insert into hash_table!\n");
				  ret = 1; /* Failed */
				  /* TODO FIX ME: handle mem failure - break from while etc... */
			  }
			  
			  kfree(tmpl3[i]);
			  tmpl3[i] = NULL;
			  i++;
		  }
	  }
	  /* Ok, inheritance and binary restrictions are handled, now lets assign the
		 level2 restrictions to p's level2 filed and place the new security in
		 p->security
	  */
	  security->level2 = level2;

  }

  p->security = security;  
  return ret;
}

static void umb_task_free_security (struct task_struct *p) {
	struct security_struct *sec = p->security;

	security_struct_destroy(p->security);
	p->security = NULL;
	kfree(sec);
}

static int umb_task_kill(struct task_struct *p, struct siginfo *info, int sig) {
	struct security_struct *security;
	int decision = 0;

	/* sanity check */
	if(current->security != NULL && p != NULL && info != NULL) {
		security = current->security;
		
        switch (sig) {

		case SIGKILL:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGKILL);
			}
			break;
			
		case SIGQUIT:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGQUIT);
			}
			break;

		case SIGTERM:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGTERM);
			}
			break;

/* 		case SIGHUP: */
/* 			if (info->si_pid != p->pid) { */
/* 				decision = bv_testbit(security->level1, NOSIGHUP); */
/* 			} */
/* 			break; */
		case SIGTRAP:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGTRAP);
			}
			break;
		case SIGCHLD:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGCHLD);
			}
			break;
		case SIGALRM:
			if (info->si_pid != p->pid) {
				decision = bv_testbit(security->level1, NOSIGALRM);
			}
			break;
 		}
        if (decision == 1 || decision == -1)
          decision = -EPERM;

        return decision;
	}
	return 0;
}

	
/***************** Registration of hooks *******************************/

static struct security_operations umbrella_security_ops = {
	/* Use the capability functions for some of the hooks */
#ifdef CONFIG_SECURITY_NETWORK
	.socket_create = umb_socket_create, 
#endif 
	.inode_mkdir =                  umb_inode_mkdir,
 	.inode_unlink =                 umb_inode_unlink,
 	.inode_link =                   umb_inode_link,
  	.inode_create =                 umb_inode_create,
   	.inode_rename =                 umb_inode_rename,
 	.inode_permission =             umb_inode_permission,
	.inode_setattr =                umb_inode_setattr,
 	.task_alloc_security =          umb_task_alloc_security,
 	.task_free_security =           umb_task_free_security,
 	.task_create =                  umb_task_create,
	.bprm_check_security =          umb_bprm_check_security,
	/* DANGEROUS - your system WILL crash enabling task_kill */
	/*  	.task_kill =                    umb_task_kill  */ 
};

/******************* Module specific code *******************************/

static int __init umbrella_init (void)
{
	/* Create procfs interface /proc/umbrella */
	umb_proc_entry = create_proc_entry("umbrella", 0222, NULL);
	if(!umb_proc_entry) {
		remove_proc_entry("umbrella", &proc_root);
		printk(KERN_INFO
			   "Umbrella: Failed creating the procfs entry.\n");
		return (-ENOMEM);
	}
	else {
		umb_proc_entry->owner = THIS_MODULE;
		umb_proc_entry->mode = S_IFREG | S_IWUGO;
		umb_proc_entry->uid = 0;
		umb_proc_entry->gid = 0;
		umb_proc_entry->size = 0;
		umb_proc_entry->write_proc = &umb_proc_write;
		
		printk(KERN_INFO "Umbrella: procfs entry created.\n");
	}



	
	/* register ourselves with the security framework */
	if (register_security(&umbrella_security_ops)) {
		printk(KERN_INFO 
			   "Failure registering Umbrella module with the kernel\n");
		/* try registering with primary module */
		if (mod_reg_security(MY_NAME, &umbrella_security_ops)) {
			printk(KERN_INFO "Failure registering Umbrella "
				" module with primary security module.\n");
			return -EINVAL;
		}
		secondary = 1;
	}
	printk(KERN_INFO "Umbrella module initialized\n");
	
	return 0;
}

static void __exit umbrella_exit (void)
{
	/* remove ourselves from the security framework */
	if (secondary) {
		if (mod_unreg_security(MY_NAME, &umbrella_security_ops))
			printk (KERN_INFO "Failure unregistering Umbrella "
				" module with primary module.\n");
	} else { 
		if (unregister_security(&umbrella_security_ops)) {
			printk(KERN_INFO "Failure unregistering Umbrella module with "
                             "the kernel\n");
		}
	}
	printk(KERN_INFO "Umbrella module removed\n");
}

security_initcall(umbrella_init);
module_exit(umbrella_exit);
